# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Plugin module which creates a rotating set of gzipped csv output files """

import os

import logging
import time
import traceback

import threading

import glob
import datetime
import pickle
import csv
import subprocess

import filter_ingest

from foglamp.common import logger

# local logger
_LOGGER = logger.setup(__name__, level=logging.INFO)


__author__ = "Douglas Orr and Deepanshu Yadav"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


FOGLAMP_DATA_ENV="FOGLAMP_DATA"
FOGLAMP_DATA_PFX="FOGLAMP_DATA/"
FOGLAMP_DEFAULT_DATA_DIR="/usr/local/foglamp/data"


_DEFAULT_CONFIG = {
    'plugin': {        # mandatory  filter
        'description': 'filter which writes its output to a series of csv files',
        'type': 'string',
        'default': 'csv_writer',
        'readonly': 'true'
    },

    'enable': {    # recommended filter
        'description': 'Enable csv writer plugin',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'order': "1"
    },
    'destDir': {
        'description': 'Destination directory',
        'type': 'string',
        'default': '{}readings-out'.format(FOGLAMP_DATA_PFX),
        'displayName': 'Destination directory',
        'order': '2'
    },
    'filterName': {
        'description': 'Value under this field will create a sub-directory inside'
                       'the Destination Directory. For example '
                       'Use south-database if the filter is applied between '
                       'south service and database',
        'type': 'string',
        'default': 'south-database',
        'displayName': 'Filter name (source-destination)',
        'order': '3'
    },
    'forwardData': {
        'description': 'forward raw data independent of filter',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Forward data even when enabled',
        'order': "4"
    },
    'addTimestamp': {
        'description': 'Add timestamp to csv data',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Add timestamp to csv data',
        'order': "5"
    },
    # csv writer-specfic values
    'maxRecords': {
        'description': 'Max records written per file',
        'type': 'integer',
        'default': '100000',
        'displayName': 'Max records',
        'order': '6'
    },
    'maxFiles': {
        'description': 'Max files accumulated before rotation',
        'type': 'integer',
        'default': '100',
        'displayName': 'Max rotation',
        'order': '7'
    },
    'fileType': {
        'description': 'Type of file in which readings to be stored.',
        'type': 'enumeration',
        'default': 'csv',
        'options': ['csv', 'pickle'],
        'displayName': 'File Type',
        'order': '8'
    },
    'enableCompress': {
        'description': 'Whether the csv files collected are to be compressed or not.',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Enable Compression for csv files.',
        'validity': " fileType == \"csv\"",
        'order': '9'
    },
    'compressionType': {
        'description': 'Compress output files',
        'type': 'enumeration',
        'default': 'bzip2',
        'options': ['bzip2', 'gzip'],
        'displayName': 'Compression Type',
        'validity': " fileType == \"csv\"  && enableCompress == \"true\"",
        'order': '10'
    },
    'assetList': {
        'description': 'Comma separated names of assets.',
        'type': 'string',
        'default': 'asset_1,asset_2',
        'validity': " fileType == \"pickle\"",
        'displayName': 'Asset List',
        'order': '11'
    },
}


class ManagementThread(threading.Thread):
    """ ManagementThread - compress / remove extra files as a background process """

    def __init__(self, args):
        super().__init__(daemon=True)  # establish thread
        _LOGGER.info("MANAGEMENT thread started")

        self.parent = args[0]  # "parent"

    def run(self):
        p = self.parent

        while True:
            p.work_cond.acquire()
            _LOGGER.info("MANAGEMENT thread: sleep")

            # check for new work
            to_compress = p.to_compress
            p.to_compress = []
            if len(to_compress) == 0:
                # release the lock and do the queued work
                p.work_cond.wait()
                _LOGGER.debug("MANAGEMENT thread: wake")

            else:
                _LOGGER.debug("MANAGEMENT thread: found {} to compress".format(to_compress))
                p.work_cond.release()

                self._compress(to_compress)

                p.work_cond.acquire()

            # ---lock held ----

            # compression done, see if we need to rotate any files
            _LOGGER.debug("MANAGEMENT thread: about to rotate")
            try:
                self._rotate()
            except:
                _LOGGER.exception("MANAGEMENT thread: rotation failed")
            finally:
                p.work_cond.release()

    def _compress(self, paths):

        p = self.parent
        compression_type = p.compress_type

        if compression_type == 'bzip2':
            # gzip all available csv's; wait for bzip to finish; do newest
            cmd = ['bzip2', '-f'] + sorted(paths, reverse=True)

        if compression_type == 'gzip':
            # gzip all available csv's; wait for bzip to finish; do newest
            cmd = ['gzip', '-f'] + sorted(paths, reverse=True)

        _LOGGER.info("COMPRESSING {}".format(cmd))
        try:
            p = subprocess.run(cmd, stderr=subprocess.PIPE)
            if p.returncode != 0:
                _err = p.stderr.decode('utf-8').strip()
                # if there was error output, log it
                _LOGGER.error("compress: runtime error")
                for line in _err.split('\n'):
                    _LOGGER.error(line)
        except:
            _LOGGER.exception("compress failure")

    def _rotate(self):
        """ _rotate - if we have more than max_files (compressed or not) remove the extra """
        p = self.parent

        to_rotate = sorted(glob.glob("{}/*.{}*".format(p.dest_dir, p.suffix)))
        while len(to_rotate) >= p.max_files:
            _LOGGER.debug("MANAGEMENT thread: found {} to rotate {}".format(to_rotate, type(to_rotate)))
            try:
                # if we have the max files possible, rotate
                _rm_file = to_rotate.pop(0)
                _LOGGER.debug("MANAGEMENT - ROTATE: removing {}".format(_rm_file))
                os.remove(_rm_file)
            except:
                traceback.exc_print()


class MediaHandler(threading.Thread):
    """
    Stores state passed to/from FogLAMP callbacks to manage periodic forwarding of data
    """

    def __init__(self, suffix, dest_dir, max_records,
                 max_files, enableCompress, compress_type, add_timestamp, filter_name):
        super().__init__()  # initialize thread class

        self.add_timestamp = add_timestamp
        if dest_dir.startswith(FOGLAMP_DATA_PFX):
            _foglamp_data_dir = os.getenv(FOGLAMP_DATA_ENV, default=FOGLAMP_DEFAULT_DATA_DIR)
            self.dest_dir = os.path.join(_foglamp_data_dir, dest_dir[len(FOGLAMP_DATA_PFX):], filter_name)
        else:
            self.dest_dir = os.path.join(dest_dir, filter_name)

        if not os.path.exists(self.dest_dir):
            os.makedirs(self.dest_dir)

        self.max_records = max_records
        self.max_files = max_files
        self.enableCompress = enableCompress
        self.compress_type = compress_type
        self.suffix = suffix

        # ### state ###
        
        self.n_records = self.max_records  # force opening a new file
        self.ext = 0
        self.cur_file_name = None
        self.media_file = None

        self._clean_dir()

        self.n_debug = 0
        self.min_debug = 100000  # for periodic debug output on common operations

        # ### management thread ### #
        self.work_cond = threading.Condition()
        self.to_compress = []  # queue of files that need to be compressed

        self.file_work_thread = ManagementThread(args=(self,))
        self.file_work_thread.start()

    def _clean_dir(self):
        """ clean_dir - remove all data files in the csv target area """
        for f in glob.glob("{}/*".format(self.dest_dir)):
            try:
                _LOGGER.debug("remove: {}".format(f))
                # preserve files with extension .csv or .csv.gz or .csv.bz2 if other than that
                # delete them
                if not f.endswith(".csv") and not f.endswith(".csv.gz") and not f.endswith(".csv.bz2"):
                    os.remove("{}".format(f))
            except:
                _LOGGER.exception("failed removing {}".format(f))

    def new_media_file(self):
        """ rotate_dir - delete the oldest file and open a new one; spawn compression process """
        _LOGGER.info("NEW MEDIA FILE")
        self.work_cond.acquire()
        try:
            _hms = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

            _cur_file = os.path.join(self.dest_dir, "{}-0000.{}".format(_hms, self.suffix))
            _LOGGER.debug("NEW - new file {}".format(_cur_file))
            if os.path.exists(_cur_file):
                # rotating in under a second -- uniquify it
                self.ext += 1
                _cur_file = os.path.join(self.dest_dir, "{}-{:04}.{}".format(_hms, self.ext, self.suffix))
                _LOGGER.debug("NEW - overlap iterate file {}".format(_cur_file))
            else:
                self.ext = 0

            self.work_cond.notify()  # look for csvs to rotate
        except:
            raise
        finally:
            self.work_cond.release()

        _LOGGER.debug("NEW: created {}".format(_cur_file))

        self._close_cur_file()

        self.media_file = self.open_media_file(_cur_file)
        self.n_records = 0
        self.cur_file_name = _cur_file

    def _close_cur_file(self):
        if self.media_file is not None:
            self.media_file = None

        if self.enableCompress and (self.cur_file_name is not None):
            # queue the previous output file for compression by the compression thread
            self.work_cond.acquire()
            try:
                if self.n_records >= self.max_records:
                    self.to_compress.append(self.cur_file_name)
                self.cur_file_name = None

                self.work_cond.notify()
            finally:
                self.work_cond.release()

    def save(self, data_iter):
        if data_iter == []:
            # filter is finished. force close
            if self.media_file:
                del self.media_file
                self.media_file = None
                self.n_records = 0
                return

        _list =[]

        # write out the generated data in chunks
        for d in data_iter:
            if self.n_records == self.max_records:
                self.new_media_file()
                        
            _list.append(d)
            if len(_list) == (self.max_records-self.n_records):
                self.write_rows(_list)
                _list = []
                        
        if _list != []:
            self.write_rows(_list)

    
class CSVHandler(MediaHandler):
    """
    Stores state passed to/from FogLAMP callbacks to manage periodic forwarding of data
    """

    def __init__(self, dest_dir, max_records,
                 max_files, enableCompress, compress_type, add_timestamp, filter_name):
        super().__init__('csv',
                         dest_dir, max_records,
                         max_files, enableCompress, compress_type, add_timestamp, filter_name                         
        )  # initialize common handler
        self.fields = None

    def open_media_file(self, cur_file):
        _media_file = csv.writer(open(cur_file, 'w'))
        return _media_file
        
    def write_rows(self, obj):
        if self.n_records == 0:
            # write the header row
            _ts = ['user_ts'] if self.add_timestamp else []
            self.media_file.writerow(self.fields + _ts)
            
        self.media_file.writerows(obj)
        self.n_records += len(obj)

    def extract_rows(self, data):
        # extract_rows - convert a list of dict's with readings into a series of ordered rows
        if data == []:
            return []

        if self.fields is None:
            # very first time through, make a note of our fields
            self.fields = list(data[0]['readings'].keys())
            self.ts_col = 'timestamp' if 'timestamp' in data[0] else ('ts' if 'ts' in data[0] else None)

        # return a row of data in a standard field-order
        for d in data:
            _readings = d['readings']

            # if we're adding a timestamp field, must not be in the data -- take it from the bodybbbbbbbbbb
            _ts = [d[self.ts_col]] if self.add_timestamp else []

            if type(_readings) is list:
                # list of dicts
                for r in _readings:
                    yield [r[field] for field in self.fields] + _ts
            else:
                # simple dict
                yield [_readings[field] for field in self.fields] + _ts


class PickleHandler(MediaHandler):
    def __init__(self, dest_dir, max_records, max_files, add_timestamp, filter_name, asset_list):
        super().__init__('pkl',
                         dest_dir, max_records,
                         max_files, False, None, add_timestamp, filter_name
        )  # initialize common handler

        # TODO handle asset_list
        self.asset_list = asset_list.split(",")

    def open_media_file(self, cur_file):
        _media_file = open(cur_file, 'wb')
        return _media_file

    def extract_rows(self, data):
        for d in data:
            yield d

    def write_rows(self, obj):
        pickle.dump(obj, self.media_file, pickle.HIGHEST_PROTOCOL)
        self.n_records += len(obj)

        
def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.info("CSVW - plugin_info called")
    return {
        'name': 'CSVWriter',
        'version': '1.9.0',
        'mode': 'none',
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the plugin configuration category
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """

    _LOGGER.info("CSVW - plugin_init called")
    try:
        _config = config  # copy.deepcopy(config)

        _dest_dir = _config['destDir']['value']
        _max_records = int(_config['maxRecords']['value'])
        _max_files =int(_config['maxFiles']['value'])
        _compress_type = _config['compressionType']['value']
        _enableCompress = _config['enableCompress']['value'] == 'true'
        _add_timestamp = _config['addTimestamp']['value'] == 'true'
        _filter_name = _config['filterName']['value']
        _asset_list = _config['assetList']['value']

        # store configuration options and working state

        _file_type = _config['fileType']['value']  # get the file type csv or pickle.
        if _file_type == 'csv':
            _config['media_handle'] = CSVHandler(
                                    dest_dir=_dest_dir,
                                    max_records=_max_records,
                                    max_files=_max_files,
                                    enableCompress=_enableCompress,
                                    compress_type=_compress_type,
                                    add_timestamp=_add_timestamp,
                                    filter_name=_filter_name)
        elif _file_type == 'pickle':
            _config['media_handle'] = PickleHandler(
                dest_dir=_dest_dir,
                max_records=_max_records,
                max_files=_max_files,
                add_timestamp=_add_timestamp,
                filter_name=_filter_name,
                asset_list=_asset_list
            )  # initialize the pickle handler class
        else:
            raise ValueError("Invalid file type {}".format(_file_type))

        _config['ingest_ref'] = ingest_ref
        _config['callback'] = callback

    except:
        _LOGGER.info("could not create configuration")
        raise

    return _config


def plugin_ingest(handle, data):
    """ plugin_ingest -- log data we receive """

    try:
        if type(data) == dict:
            # make everything a list
            data = [data]

        filter_enabled = handle['enable']['value'] == 'true'
        allow_raw_data = handle['forwardData']['value'] == 'true'
        media_handle = handle['media_handle']

        if filter_enabled:
            media_handle.save(media_handle.extract_rows(data))

        if not filter_enabled or allow_raw_data:
            # pass data upstream if not filtering, or if explicit forward_data flagged
            filter_ingest.filter_ingest_callback(handle['callback'], handle['ingest_ref'], data)

    except Exception as ex:
        _LOGGER.error("csv writer exception {}".format(traceback.format_exc()))
        raise


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _LOGGER.info("Old config for  plugin {} \n new config {}".format(handle, new_config))
    plugin_shutdown(handle)

    # plugin_init
    new_handle = plugin_init(new_config, handle['ingest_ref'], handle['callback'])

    return new_handle


def plugin_shutdown(handle):
    """ Shut down the plugin.
    Args:
        handle: JSON configuration document for the plugin configuration category
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """

    _media_handle = handle['media_handle']

    if _media_handle is not None:
        _media_handle._close_cur_file()
        time.sleep(5)  # let the compression get fully under way
    _media_handle = None
